import datetime
import pydantic
import rframe
from pydantic import validator
from typing import List, Literal

from ..base_schemas import XeDoc
from ..._settings import settings

from ..constants import SOURCE


class BaseCalibration(XeDoc):
    """Base class for calibration metadata"""

    _ALIAS = ""
    _CATEGORY = "calibration"

    class Config:
        allow_population_by_field_name = True

    time: rframe.Interval[datetime.datetime] = rframe.IntervalIndex(alias="run_id")
    source_id: str = rframe.Index(min_length=1, max_length=60)

    source_type: SOURCE

    operator: str = pydantic.Field(min_length=1, max_length=60, default="")
    comments: str = ""
    run_ids: List[str] = []
    filled_by: str = pydantic.Field(min_length=1, max_length=60, default="")

    @validator("time", pre=True)
    def run_id_to_time(cls, v):
        """Convert run id to time"""
        if isinstance(v, (str, int)):
            try:
                v = settings.run_id_to_interval(v)
            except:
                pass
        return v



from collections import defaultdict
import os
import json
import xedocs
import logging
import fsspec

from pathlib import Path
from tinydb import TinyDB
from pydantic import BaseSettings
from tinydb.storages import Storage
from typing import Any, Dict, Optional
from xedocs.database_interface import DatabaseInterface


XEDOCS_LOCAL_REPO_ENV = os.getenv(
    "XEDOCS_LOCAL_REPO_ENV", os.path.join(xedocs.settings.CONFIG_DIR, "local_repo.env")
)
logger = logging.getLogger(__name__)


class FsspecStorage(Storage):
    def __init__(self, path, chunksize=1000, indent=4, **storage_kwargs):
        self.path = path
        self.chunksize = chunksize
        self.indent = indent
        self.storage_kwargs = storage_kwargs

    def read(self) -> Optional[Dict[str, Dict[str, Any]]]:
        datasets = {}
        with fsspec.open_files(self.path, **self.storage_kwargs) as fs:
            for f in fs:
                if not f.size:
                    continue
                if not f.readable():
                    continue
                filedata = json.load(f)
                for table, data in filedata.items():
                    if table in datasets:
                        datasets[table].update(data)
                    else:
                        datasets[table] = data

        if not datasets:
            return None
        return datasets

    def write(self, data: Dict[str, Dict[str, Any]]) -> None:
        filenum = 1
        cnt = 0
        docs = defaultdict(dict)
        for name, table in data.items():
            for num, doc in table.items():
                docs[name][num] = doc
                cnt += 1
                if sum([len(d) for d in docs]) > self.chunksize:
                    self._write_chunk(docs, filenum)
                    filenum += 1
                    docs = defaultdict(dict)
        if sum([len(d) for d in docs]):
            self._write_chunk(docs, filenum)

    def _write_chunk(self, data, filenum):
        path = os.path.join(self.path, f"{filenum}.json")
        with fsspec.open(path, "wb", **self.storage_kwargs) as f:
            if not f.writable():
                raise IOError(
                    f'Cannot write to the database. Access mode is "{f.mode}"'
                )
            json.dump(data, f, indent=self.indent)


class LocalRepoSettings(BaseSettings):
    class Config:
        env_prefix = "XEDOCS_LOCAL_REPO_"
        env_file = XEDOCS_LOCAL_REPO_ENV

    PRIORITY: int = 3
    PATH: str = xedocs.settings.DATA_DIR


class LocalRepoDatabase(DatabaseInterface):
    settings: LocalRepoSettings = LocalRepoSettings()
    database: str
    alias: str

    def __init__(
        self,
        database: str = None,
        alias: str = None,
        settings: LocalRepoSettings = None,
    ):
        self.database = database
        if alias is None:
            alias = database
        self.alias = alias
        if settings is None:
            settings = LocalRepoSettings()
        self.settings = settings

    def base_path_for_schema(self, schema):
        return (
            Path(self.settings.PATH) / self.database / schema._CATEGORY / schema._ALIAS
        )

    def datasource_for_schema(self, schema):
        basepath = self.base_path_for_schema(schema)
        basepath = basepath.expanduser().resolve()
        if not basepath.exists():
            basepath.mkdir(parents=True)
        path = basepath / "*.json"
        db = TinyDB(path.absolute(), storage=FsspecStorage)
        return db.table(schema._ALIAS)



from collections import defaultdict


from rframe import DataAccessor
from xedocs import all_schemas, schemas_by_category


class XedocsContext:
    _schemas: dict
    _overrides: dict
    _db_name: str

    def __init__(self, schemas: dict = None, db_name=None, overrides=None):

        self._db_name = db_name

        if schemas is None:
            schemas = all_schemas()
        self._schemas = schemas

        if overrides is None:
            overrides = {}
        self._overrides = overrides

    def get_accessor(self, schema, name=None):
        if name is None:
            name = schema._ALIAS
        if name in self._overrides:
            return DataAccessor(schema, self._overrides[name])
        return getattr(schema, self._db_name, None)

    def __getitem__(self, key):
        schema = self._schemas.get(key, None)
        if schema is None:
            raise KeyError(f"No schema named {key} found")

        if isinstance(schema, dict):
            return XedocsContext(
                schema, db_name=self._db_name, overrides=self._overrides
            )

        accessor = self.get_accessor(schema, name=key)

        if accessor is None:
            raise KeyError(f"No accessor named {key} found")

        return accessor

    def __getattr__(self, attr):
        if attr in self._schemas.keys():
            return self[attr]
        raise AttributeError(attr)

    def __dir__(self):
        return super().__dir__() + list(self._schemas.keys())


def straxen_db(schemas=None, datasource_overrides=None, by_category=False):
    if schemas is None:
        if by_category:
            schemas = schemas_by_category()
        else:
            schemas = all_schemas()

    return XedocsContext(schemas, db_name="straxen_db", overrides=datasource_overrides)


def analyst_db(schemas=None, datasource_overrides=None, by_category=False):
    if schemas is None:
        if by_category:
            schemas = schemas_by_category()
        else:
            schemas = all_schemas()

    return XedocsContext(schemas, db_name="analyst_db", overrides=datasource_overrides)



"""Top-level package for xedocs."""

__author__ = """Yossi Mosbacher"""
__email__ = "joe.mosbacher@gmail.com"
__version__ = "0.2.10"


import logging

logger = logging.getLogger(__name__)


from ._settings import settings
from . import utils

from . import schemas
from . import xedocs
from .xedocs import *
from . import api
from . import database_interface

__all__ = [
    "settings",
    "utils",
    "schemas",
] + xedocs.__all__

try:
    from . import _straxen_plugin

except ImportError:
    logger.warning(
        "Could not register straxen protocol, \
                     most likely straxen not installed."
    )

except ValueError:
    pass

try:
    from . import widgets

    gui = widgets.XedocsEditor()
except ImportError:
    logger.warning("Could not import editors, GUI not available.")

try:
    from .entrypoints import load_entry_points

    hooks = load_entry_points("xedocs")

    for name, hook in hooks.items():
        if callable(hook):
            hook()
except Exception as e:
    logger.warning(f"Could not load plugins. Failed with error: {e}")

try:
    from .database_interface import load_db_interfaces

    load_db_interfaces()

except Exception as e:
    logger.warning(f"Could not load database interfaces. Failed with error: {e}")


for schema in xedocs.all_schemas().values():
    try:
        settings.register_databases(schema)
    except Exception as e:
        logger.warning(
            f"Could not load databases for {schema._ALIAS}. Failed with error: {e}"
        )


def __getattr__(name: str):
    if name in settings.DATABASES:
        interfaces = settings.database_interfaces(name)
        return utils.DatasetCollection(interfaces)
    raise AttributeError(name)


def __dir__():
    return list(settings.DATABASES) + __all__



import os
import xedocs
import tinydb
import logging

from pydantic import BaseSettings
from xedocs.schemas import XeDoc
from xedocs.database_interface import DatabaseInterface

from .local import FsspecStorage

XEDOCS_GITHUB_ENV = os.getenv(
    "XEDOCS_GITHUB_ENV", os.path.join(xedocs.settings.CONFIG_DIR, "github.env")
)
logger = logging.getLogger(__name__)


def default_github_username():
    try:
        from git import config

        cfg_path = config.get_config_path("global")
        cfg = config.GitConfigParser(cfg_path)
        return cfg.get("github", "user")
    except:
        return None


def default_github_token():
    try:
        from git import config

        cfg_path = config.get_config_path("global")
        cfg = config.GitConfigParser(cfg_path)
        return cfg.get("github", "token")
    except:
        return None


class GithubSettings(BaseSettings):
    class Config:
        env_prefix = "XEDOCS_GITHUB_"
        env_file = XEDOCS_GITHUB_ENV

    PRIORITY: int = -1
    ORG: str = "XENONnT"
    REPO: str = "xedocs-data"
    URL_TEMPLATE: str = "github://{org}:{repo}@/{database}/{category}/{name}/*.json"
    USERNAME: str = default_github_username()
    TOKEN: str = default_github_token()


class GithubDatabase(DatabaseInterface):
    settings: GithubSettings = GithubSettings()

    def __init__(
        self, database: str = None, alias: str = None, settings: GithubSettings = None
    ):
        self.database = database
        if alias is None:
            alias = database
        self.alias = alias
        if settings is not None:
            self.settings = settings

    def github_url_for_schema(self, schema: XeDoc):
        return self.settings.URL_TEMPLATE.format(
            org=self.settings.ORG,
            repo=self.settings.REPO,
            category=schema._CATEGORY,
            name=schema._ALIAS,
            database=self.database,
        )

    def datasource_for_schema(self, schema):
        url = self.github_url_for_schema(schema)
        db = tinydb.TinyDB(
            url,
            storage=FsspecStorage,
            username=self.settings.USERNAME,
            token=self.settings.TOKEN,
        )
        return db.table(schema._ALIAS)
